# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from dataclasses import dataclass

from quesadilla.core.task_executables import (
    AsyncTaskExecutable,
    SyncTaskExecutable,
    TaskExecutable,
)
from quesadilla.core.task_queues import TaskQueue


@dataclass(frozen=True)
class TaskDefinition[**P, T]:
    task_queue: TaskQueue
    task_executable: TaskExecutable[P, T]


@dataclass(frozen=True)
class SyncTaskDefinition[**P, T](TaskDefinition[P, T]):
    task_executable: SyncTaskExecutable[P, T]

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self.task_executable(*args, **kwargs)


@dataclass(frozen=True)
class AsyncTaskDefinition[**P, T](TaskDefinition[P, T]):
    task_executable: AsyncTaskExecutable[P, T]

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return await self.task_executable(*args, **kwargs)



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import uuid
from typing import Any

import ulid

from quesadilla.core.task_serialization import SerializedTask


class TaskJSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, (ulid.ULID, uuid.UUID)):  # pragma: nobranch
            return str(o)
        return super().default(o)  # pragma: nocover


class TaskJSONDecoder(json.JSONDecoder):
    def __init__(self, **kwargs: Any):
        kwargs["object_hook"] = self.object_hook
        super().__init__(**kwargs)

    def object_hook(self, obj: Any) -> Any:  # type: ignore
        if isinstance(obj, (list, tuple)):
            return tuple(map(self.object_hook, obj))  # type: ignore

        if isinstance(obj, dict):
            return {k: self.object_hook(v) for k, v in obj.items()}  # type: ignore

        if isinstance(obj, str):
            try:
                return ulid.ULID.from_str(obj)  # type: ignore
            except ValueError:
                pass
            try:
                return uuid.UUID(obj)  # type: ignore
            except ValueError:
                pass

        return obj


class JSONConnectorMixin:
    @staticmethod
    def encode(o: Any) -> str:
        return json.dumps(o, cls=TaskJSONEncoder)

    @staticmethod
    def decode(o: str) -> SerializedTask[Any]:
        return json.loads(o, cls=TaskJSONDecoder)



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections.abc import MutableMapping
from dataclasses import dataclass, field
from typing import Any

import ulid

from quesadilla.connectors.basic_queues import BaseQueue, BasicQueue
from quesadilla.connectors.mixins import JSONConnectorMixin
from quesadilla.connectors.protocol import Connector
from quesadilla.core.errors import TaskAlreadyQueued
from quesadilla.core.task_serialization import SerializedTask

# this in-memory connector is designed for use in testing and prototyping

BufferEntryType = BasicQueue[ulid.ULID]
BufferQueueType = MutableMapping[str, BufferEntryType]
BufferType = MutableMapping[str, BufferQueueType]
LedgerEntryType = MutableMapping[ulid.ULID, str]
LedgerQueueType = MutableMapping[str, LedgerEntryType]
LedgerType = MutableMapping[str, LedgerQueueType]


@dataclass(frozen=True)
class InMemoryConnector(JSONConnectorMixin, Connector):
    buffer: BufferType = field(init=False, repr=False)
    ledger: LedgerType = field(init=False, repr=False)

    def queue(self, serialized: SerializedTask[Any]) -> None:
        if (
            serialized["id"]
            in self.ledger[serialized["namespace"]][serialized["task_queue_name"]]
        ):
            raise TaskAlreadyQueued(
                namespace=serialized["namespace"],
                task_queue_name=serialized["task_queue_name"],
                task_name=serialized["task_name"],
                task_id=serialized["id"],
            )

        self.ledger[serialized["namespace"]][serialized["task_queue_name"]][
            serialized["id"]
        ] = self.encode(serialized)
        self.buffer[serialized["namespace"]][serialized["task_queue_name"]].put(
            serialized["id"]
        )

    async def aqueue(self, serialized: SerializedTask[Any]) -> None:
        return self.queue(serialized)

    def find(
        self, namespace: str, task_queue: str, task_id: ulid.ULID
    ) -> SerializedTask[Any] | None:
        if (qt := self.ledger[namespace][task_queue].get(task_id)) is None:
            return None
        return self.decode(qt)

    async def afind(
        self, namespace: str, task_queue: str, task_id: ulid.ULID
    ) -> SerializedTask[Any] | None:
        return self.find(namespace, task_queue, task_id)

    def pull(self, namespace: str, task_queue: str) -> SerializedTask[Any] | None:
        try:
            task_id = self.buffer[namespace][task_queue].get(timeout=0.1)
            return self.decode(self.ledger[namespace][task_queue][task_id])
        except (BaseQueue.Empty, KeyError):
            return None

    def update(
        self,
        namespace: str,
        task_queue: str,
        serialized: SerializedTask[Any],
    ) -> None:
        self.ledger[namespace][task_queue][serialized["id"]] = self.encode(serialized)



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from functools import cached_property
from typing import Any

import ulid

from quesadilla.core.task_definitions import (
    AsyncTaskDefinition,
    SyncTaskDefinition,
    TaskDefinition,
)
from quesadilla.core.task_executables import (
    AsyncTaskExecutable,
    Failure,
    Success,
    SyncTaskExecutable,
)
from quesadilla.core.task_metadata import BaseTaskMetadata, TaskExecutionContext
from quesadilla.core.task_queues import GenericQueuedTask, TaskQueue


@dataclass(frozen=True)
class Task[**P, T]:
    task_definition: TaskDefinition[P, T]

    @cached_property
    def task_queue(self) -> TaskQueue:
        return self.task_definition.task_queue

    def upgrade(self, queued_task: GenericQueuedTask) -> QueuedTask[T]:
        return QueuedTask.from_generic(queued_task, self)

    def queue(self, *args: P.args, **kwargs: P.kwargs) -> QueuedTask[T]:
        task_execution_context = TaskExecutionContext(args=tuple(args), kwargs=kwargs)
        queued_task = QueuedTask(task=self, execution_context=task_execution_context)

        self.task_queue.queue(queued_task.generic)

        return queued_task

    async def aqueue(self, *args: P.args, **kwargs: P.kwargs) -> QueuedTask[T]:
        task_execution_context = TaskExecutionContext(args=tuple(args), kwargs=kwargs)
        queued_task = QueuedTask(task=self, execution_context=task_execution_context)

        await self.task_queue.aqueue(queued_task.generic)

        return queued_task

    def find(self, task_id: ulid.ULID) -> QueuedTask[T] | None:
        queued_task: QueuedTask[T] | None = (
            QueuedTask.from_generic(generic_queued_task, self)
            if (generic_queued_task := self.task_queue.find(task_id))
            else None
        )
        return queued_task

    async def afind(self, task_id: ulid.ULID) -> QueuedTask[T] | None:
        queued_task: QueuedTask[T] | None = (
            QueuedTask.from_generic(generic_queued_task, self)
            if (generic_queued_task := await self.task_queue.afind(task_id))
            else None
        )
        return queued_task


@dataclass(frozen=True)
class SyncTask[**P, T](Task[P, T]):
    task_definition: SyncTaskDefinition[P, T]

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self.task_definition(*args, **kwargs)


def sync_task[
    **P, T
](
    task_queue: TaskQueue,
    *,
    name: str | None = None,
) -> Callable[
    [Callable[P, T]], SyncTask[P, T]
]:
    def wrapper(f: Callable[P, T]) -> SyncTask[P, T]:
        task_executable = SyncTaskExecutable(name or f.__name__, f)
        task_queue.register_task_executable(task_executable)
        return SyncTask(SyncTaskDefinition(task_queue, task_executable))

    return wrapper


@dataclass(frozen=True)
class AsyncTask[**P, T](Task[P, T]):
    task_definition: AsyncTaskDefinition[P, T]

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return await self.task_definition(*args, **kwargs)


def async_task[
    **P, T
](
    task_queue: TaskQueue,
    *,
    name: str | None = None,
) -> Callable[
    [Callable[P, Awaitable[T]]], AsyncTask[P, T]
]:
    def wrapper(f: Callable[P, Awaitable[T]]) -> AsyncTask[P, T]:
        task_executable = AsyncTaskExecutable(name or f.__name__, f)
        task_queue.register_task_executable(task_executable)
        return AsyncTask(AsyncTaskDefinition(task_queue, task_executable))

    return wrapper


@dataclass(frozen=True)
class QueuedTask[T](BaseTaskMetadata[T]):
    task: Task[Any, T] = field(kw_only=True)

    @cached_property
    def generic(self) -> GenericQueuedTask:
        return GenericQueuedTask(
            id=self.id,
            task_queue=self.task.task_definition.task_queue,
            task_executable=self.task.task_definition.task_executable,
            execution_context=self.execution_context,
            execution_result=self.execution_result,
            queued_at=self.queued_at,
            started_at=self.started_at,
            finalized_at=self.finalized_at,
        )

    @staticmethod
    def from_generic[
        NT
    ](queued_task: GenericQueuedTask, task: Task[Any, NT]) -> QueuedTask[NT]:
        return QueuedTask(
            id=queued_task.id,
            execution_context=queued_task.execution_context,
            execution_result=queued_task.execution_result,
            queued_at=queued_task.queued_at,
            started_at=queued_task.started_at,
            finalized_at=queued_task.finalized_at,
            task=task,
        )

    @cached_property
    def result(self) -> Success[T] | Failure:
        return self.generic.result

    def refresh(self) -> QueuedTask[T]:
        return self.task.upgrade(self.generic.refresh())

    async def arefresh(self) -> QueuedTask[T]:
        return self.task.upgrade(await self.generic.arefresh())

    def wait_for(self, interval: float = 0.1) -> QueuedTask[T]:
        return self.task.upgrade(self.generic.wait_for(interval))

    async def await_for(self, interval: float = 0.1) -> QueuedTask[T]:
        return self.task.upgrade(await self.generic.await_for(interval))



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass

import ulid


class QuesadillaException(Exception):
    def __str__(self) -> str:
        return ""


@dataclass()
class TaskQueueException(QuesadillaException):
    namespace: str
    task_queue_name: str

    def __str__(self) -> str:
        return (
            f"namespace={repr(self.namespace)}, "
            f"task_queue_name={repr(self.task_queue_name)}"
        )


@dataclass()
class UnknownTaskQueueNameError(TaskQueueException):
    pass


@dataclass()
class DuplicateTaskQueueNameError(TaskQueueException):
    pass


@dataclass()
class TaskNameException(TaskQueueException):
    task_name: str

    def __str__(self) -> str:
        return (
            f"namespace={repr(self.namespace)}, "
            f"task_queue_name={repr(self.task_queue_name)}, "
            f"task_name={repr(self.task_name)}"
        )


@dataclass()
class UnknownTaskNameError(TaskNameException):
    pass


@dataclass()
class DuplicateTaskNameError(TaskNameException):
    pass


@dataclass()
class TaskException(Exception):
    namespace: str
    task_queue_name: str
    task_name: str
    task_id: ulid.ULID

    def __str__(self) -> str:
        return (
            f"namespace={repr(self.namespace)}, "
            f"task_queue_name={repr(self.task_queue_name)}, "
            f"task_name={repr(self.task_name)}, "
            f"task_id={repr(self.task_id)}"
        )


class TaskAlreadyQueued(TaskException):
    pass


class TaskNotFinishedError(TaskException):
    pass


class TaskDoesNotExistError(TaskException):
    pass


class TaskDoesNotExistAnymoreError(TaskDoesNotExistError):
    pass



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import asyncio
import datetime
import time
from collections.abc import MutableMapping
from dataclasses import dataclass, field
from functools import cached_property
from typing import Any, Self

import ulid

from quesadilla.connectors.protocol import Connector
from quesadilla.core.errors import (
    DuplicateTaskNameError,
    DuplicateTaskQueueNameError,
    TaskDoesNotExistAnymoreError,
    TaskNotFinishedError,
    UnknownTaskNameError,
)
from quesadilla.core.logging import QueuedTaskLoggingExtras, TaskQueueLoggingExtras
from quesadilla.core.task_executables import Failure, Success, TaskExecutable
from quesadilla.core.task_metadata import BaseTaskMetadata, SerializableTask


@dataclass(frozen=True)
class TaskQueue:
    name: str
    namespace: str

    connector: Connector = field(kw_only=True, repr=False)
    task_executables: MutableMapping[str, TaskExecutable[Any, Any]] = field(
        default_factory=dict, init=False, repr=False
    )

    def register_task_executable(
        self, task_executable: TaskExecutable[Any, Any]
    ) -> None:
        if task_executable.name in self.task_executables:
            raise DuplicateTaskNameError(
                namespace=self.namespace,
                task_queue_name=self.name,
                task_name=task_executable.name,
            )
        self.task_executables[task_executable.name] = task_executable

    def get_task_executable(self, name: str) -> TaskExecutable[Any, Any]:
        try:
            return self.task_executables[name]
        except KeyError as e:
            raise UnknownTaskNameError(
                namespace=self.namespace,
                task_queue_name=self.name,
                task_name=name,
            ) from e

    def queue(self, queued_task: GenericQueuedTask) -> None:
        self.connector.queue(queued_task.serializable.serialize())

    async def aqueue(self, queued_task: GenericQueuedTask) -> None:
        await self.connector.aqueue(queued_task.serializable.serialize())

    def find(self, task_id: ulid.ULID) -> GenericQueuedTask | None:
        if (s := self.connector.find(self.namespace, self.name, task_id)) is None:
            return None

        return GenericQueuedTask.from_serializable(
            (serializable := SerializableTask.deserialize(s)),
            task_queue=self,
            task_executable=self.get_task_executable(serializable.task_name),
        )

    async def afind(self, task_id: ulid.ULID) -> GenericQueuedTask | None:
        if (
            s := await self.connector.afind(self.namespace, self.name, task_id)
        ) is None:
            return None

        return GenericQueuedTask.from_serializable(
            (serializable := SerializableTask.deserialize(s)),
            task_queue=self,
            task_executable=self.get_task_executable(serializable.task_name),
        )

    def pull(self) -> GenericQueuedTask | None:
        if (s := self.connector.pull(self.namespace, self.name)) is None:
            return None

        return GenericQueuedTask.from_serializable(
            (serializable := SerializableTask.deserialize(s)),
            task_queue=self,
            task_executable=self.get_task_executable(serializable.task_name),
        )

    def update(self, queued_task: GenericQueuedTask) -> None:
        self.connector.update(
            self.namespace, self.name, queued_task.serializable.serialize()
        )

    @cached_property
    def logging_extras(self) -> TaskQueueLoggingExtras:
        return {
            "namespace": self.namespace,
            "task_queue_name": self.name,
        }


@dataclass(frozen=True)
class TaskNamespace:
    namespace: str
    connector: Connector = field(kw_only=True)

    task_queues: MutableMapping[str, TaskQueue] = field(
        default_factory=dict, init=False, repr=False
    )

    def new(self, name: str) -> TaskQueue:
        if name in self.task_queues:
            raise DuplicateTaskQueueNameError(
                namespace=self.namespace,
                task_queue_name=name,
            )

        task_queue = TaskQueue(name, namespace=self.namespace, connector=self.connector)
        self.task_queues[name] = task_queue

        return task_queue


@dataclass(frozen=True)
class GenericQueuedTask(BaseTaskMetadata[Any]):
    task_queue: TaskQueue = field(kw_only=True)
    task_executable: TaskExecutable[Any, Any] = field(kw_only=True)

    @cached_property
    def serializable(self) -> SerializableTask[Any]:
        return SerializableTask(
            id=self.id,
            namespace=self.task_queue.namespace,
            task_queue_name=self.task_queue.name,
            task_name=self.task_executable.name,
            execution_context=self.execution_context,
            execution_result=self.execution_result,
            queued_at=self.queued_at,
            started_at=self.started_at,
            finalized_at=self.finalized_at,
        )

    @classmethod
    def from_serializable(
        cls,
        serializable: SerializableTask[Any],
        task_queue: TaskQueue,
        task_executable: TaskExecutable[Any, Any],
    ) -> Self:
        return cls(
            id=serializable.id,
            task_queue=task_queue,
            task_executable=task_executable,
            execution_context=serializable.execution_context,
            execution_result=serializable.execution_result,
            queued_at=serializable.queued_at,
            started_at=serializable.started_at,
            finalized_at=serializable.finalized_at,
        )

    @cached_property
    def result(self) -> Success[Any] | Failure:
        if not self.execution_result.finished:
            raise TaskNotFinishedError(
                namespace=self.task_queue.namespace,
                task_queue_name=self.task_queue.name,
                task_name=self.task_executable.name,
                task_id=self.id,
            )

        if self.execution_result.exc is not None:
            return Failure(self.execution_result.exc)
        return Success(self.execution_result.value)

    def refresh(self) -> GenericQueuedTask:
        if (queued_task := self.task_queue.find(self.id)) is None:
            raise TaskDoesNotExistAnymoreError(
                namespace=self.task_queue.namespace,
                task_queue_name=self.task_queue.name,
                task_name=self.task_executable.name,
                task_id=self.id,
            )
        return queued_task

    async def arefresh(self) -> GenericQueuedTask:
        if (queued_task := await self.task_queue.afind(self.id)) is None:
            raise TaskDoesNotExistAnymoreError(
                namespace=self.task_queue.namespace,
                task_queue_name=self.task_queue.name,
                task_name=self.task_executable.name,
                task_id=self.id,
            )
        return queued_task

    def wait_for(self, interval: float = 0.1) -> GenericQueuedTask:
        while not (refreshed := self.refresh()).finalized:
            time.sleep(interval)
        return refreshed

    async def await_for(self, interval: float = 0.1) -> GenericQueuedTask:
        while not (refreshed := await self.arefresh()).finalized:
            await asyncio.sleep(interval)
        return refreshed

    def start(self) -> GenericQueuedTask:
        return GenericQueuedTask(
            id=self.id,
            task_queue=self.task_queue,
            task_executable=self.task_executable,
            execution_context=self.execution_context,
            execution_result=self.execution_result,
            queued_at=self.queued_at,
            started_at=datetime.datetime.now(datetime.UTC),
            finalized_at=self.finalized_at,
        )

    def execute(self, asyncio_runner: asyncio.Runner) -> GenericQueuedTask:
        return GenericQueuedTask(
            id=self.id,
            task_queue=self.task_queue,
            task_executable=self.task_executable,
            execution_context=self.execution_context,
            execution_result=self.task_executable.execute(
                asyncio_runner,
                *self.execution_context.args,
                *self.execution_context.kwargs,
            ),
            queued_at=self.queued_at,
            started_at=self.started_at,
            finalized_at=self.finalized_at,
        )

    def finalize(self) -> GenericQueuedTask:
        return GenericQueuedTask(
            id=self.id,
            task_queue=self.task_queue,
            task_executable=self.task_executable,
            execution_context=self.execution_context,
            execution_result=self.execution_result,
            queued_at=self.queued_at,
            started_at=self.started_at,
            finalized_at=datetime.datetime.now(datetime.UTC),
        )

    def reset(self) -> GenericQueuedTask:
        return GenericQueuedTask(
            id=self.id,
            task_queue=self.task_queue,
            task_executable=self.task_executable,
            execution_context=self.execution_context,
            execution_result=self.execution_result,
            queued_at=self.queued_at,
            started_at=None,
            finalized_at=None,
        )

    @cached_property
    def logging_extras(self) -> QueuedTaskLoggingExtras:
        return {
            "task_queue": self.task_queue.logging_extras,
            "task_name": self.task_executable.name,
            "id": str(self.id),
            "finished": self.execution_result.finished,
        }



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Any, Protocol

import ulid

from quesadilla.core.task_serialization import SerializedTask


class Connector(Protocol):
    def queue(self, serialized: SerializedTask[Any]) -> None: ...  # pragma: nocover

    async def aqueue(
        self, serialized: SerializedTask[Any]
    ) -> None: ...  # pragma: nocover

    def find(
        self, namespace: str, task_queue: str, task_id: ulid.ULID
    ) -> SerializedTask[Any] | None: ...  # pragma: nocover

    async def afind(
        self, namespace: str, task_queue: str, task_id: ulid.ULID
    ) -> SerializedTask[Any] | None: ...  # pragma: nocover

    def pull(
        self, namespace: str, task_queue: str
    ) -> SerializedTask[Any] | None: ...  # pragma: nocover

    def update(
        self,
        namespace: str,
        task_queue: str,
        serialized: SerializedTask[Any],
    ) -> None: ...  # pragma: nocover



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import queue
from typing import Protocol

# this Protocol represents anything that has a FIFO-queue interface
# for internal use only, to unify some interfaces for Python builtins


class BasicQueue[T](Protocol):
    def get(self, timeout: float | None = None) -> T: ...  # pragma: nocover

    def put(self, v: T, timeout: float | None = None) -> None: ...  # pragma: nocover


class BaseQueue[T](BasicQueue[T]):
    class Empty(queue.Empty):
        pass

    class Full(queue.Full):
        pass



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class TaskExecutionContext:
    args: tuple[Any, ...]
    kwargs: dict[str, Any]


@dataclass(frozen=True)
class Success[T]:
    value: T


@dataclass(frozen=True)
class Failure:
    exc: str


@dataclass(frozen=True)
class TaskExecutionResult[T]:
    finished: bool = False
    value: T | None = None
    exc: str | None = None


@dataclass(frozen=True)
class TaskExecutable[**P, T]:
    name: str
    f: Callable[P, T] | Callable[P, Awaitable[T]]

    def execute(
        self, asyncio_runner: asyncio.Runner, *args: P.args, **kwargs: P.kwargs
    ) -> TaskExecutionResult[T]: ...  # pragma: nocover


@dataclass(frozen=True)
class SyncTaskExecutable[**P, T](TaskExecutable[P, T]):
    f: Callable[P, T]

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self.f(*args, **kwargs)

    def execute(
        self, asyncio_runner: asyncio.Runner, *args: P.args, **kwargs: P.kwargs
    ) -> TaskExecutionResult[T]:
        try:
            return TaskExecutionResult(finished=True, value=self(*args, **kwargs))
        except Exception as e:
            return TaskExecutionResult(finished=True, exc=repr(e))


@dataclass(frozen=True)
class AsyncTaskExecutable[**P, T](TaskExecutable[P, T]):
    f: Callable[P, Awaitable[T]]

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return await self.f(*args, **kwargs)

    def execute(
        self, asyncio_runner: asyncio.Runner, *args: P.args, **kwargs: P.kwargs
    ) -> TaskExecutionResult[T]:
        try:
            return TaskExecutionResult(
                finished=True, value=asyncio_runner.run(self(*args, **kwargs))
            )
        except Exception as e:
            return TaskExecutionResult(finished=True, exc=repr(e))



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import queue
from collections import defaultdict
from dataclasses import dataclass, field

from quesadilla.connectors.basic_queues import BaseQueue
from quesadilla.connectors.in_memory.base import (
    BufferType,
    InMemoryConnector,
    LedgerType,
)


@dataclass(frozen=True)
class Queue[T](BaseQueue[T]):
    q: queue.Queue[T] = field(default_factory=queue.Queue)

    def get(self, timeout: float | None = None) -> T:
        try:
            return self.q.get(block=True, timeout=timeout)
        except queue.Empty as e:
            raise self.Empty(*e.args) from e

    def put(self, v: T, timeout: float | None = None) -> None:
        self.q.put(v, block=True, timeout=timeout)


@dataclass(frozen=True)
class ThreadSafeInMemoryConnector(InMemoryConnector):
    buffer: BufferType = field(
        default_factory=lambda: defaultdict(lambda: defaultdict(Queue)),
        init=False,
        repr=False,
    )
    ledger: LedgerType = field(
        default_factory=lambda: defaultdict(lambda: defaultdict(dict)),
        init=False,
        repr=False,
    )



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime
from typing import Any, TypedDict

import ulid


def from_timestamp(timestamp: float) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(timestamp, tz=datetime.UTC)


def from_timestamp_or_none(timestamp: float | None) -> datetime.datetime | None:
    return from_timestamp(timestamp) if timestamp else None


def to_timestamp(dt: datetime.datetime) -> float:
    return dt.timestamp()


def to_timestamp_or_none(dt: datetime.datetime | None) -> float | None:
    return to_timestamp(dt) if dt else None


class SerializedTaskExecutionContext(TypedDict):
    args: tuple[Any]
    kwargs: dict[str, Any]


class SerializedTaskExecutionResult[T](TypedDict):
    finished: bool
    value: T | None
    exc: str | None


class SerializedTask[T](TypedDict):
    id: ulid.ULID
    namespace: str
    task_queue_name: str
    task_name: str
    execution_context: SerializedTaskExecutionContext
    execution_result: SerializedTaskExecutionResult[T]
    queued_at: float
    started_at: float | None
    finalized_at: float | None



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import multiprocessing.managers
import multiprocessing.queues
import queue
from dataclasses import dataclass, field
from typing import Any

from quesadilla.connectors.basic_queues import BaseQueue
from quesadilla.connectors.in_memory.base import (
    BufferType,
    InMemoryConnector,
    LedgerType,
)
from quesadilla.core.task_serialization import SerializedTask


@dataclass(frozen=True)
class Queue[T](BaseQueue[T]):
    q: multiprocessing.queues.Queue[T]

    def get(self, timeout: float | None = None) -> T:  # pragma: nocover
        # this function is excluded from coverage
        # because it's executed in a separate process
        try:
            return self.q.get(block=True, timeout=timeout)
        except queue.Empty as e:
            raise self.Empty(*e.args) from e

    def put(self, v: T, timeout: float | None = None) -> None:
        self.q.put(v, block=True, timeout=timeout)


@dataclass(frozen=True)
class ProcessSafeInMemoryConnector(InMemoryConnector):
    buffer: BufferType = field(init=False, repr=False)
    ledger: LedgerType = field(init=False, repr=False)

    object_manager: multiprocessing.managers.SyncManager = field(
        default_factory=multiprocessing.Manager,
        init=False,
        repr=False,
    )

    def __post_init__(self) -> None:
        object.__setattr__(self, "buffer", self.object_manager.dict())
        object.__setattr__(self, "ledger", self.object_manager.dict())

    def queue(self, serialized: SerializedTask[Any]) -> None:
        if not serialized["namespace"] in self.buffer:
            self.buffer[serialized["namespace"]] = self.object_manager.dict()
        if not serialized["task_queue_name"] in self.buffer[serialized["namespace"]]:
            self.buffer[serialized["namespace"]][serialized["task_queue_name"]] = Queue(
                self.object_manager.Queue()  # type: ignore
            )

        if not serialized["namespace"] in self.ledger:
            self.ledger[serialized["namespace"]] = self.object_manager.dict()
        if not serialized["task_queue_name"] in self.ledger[serialized["namespace"]]:
            self.ledger[serialized["namespace"]][
                serialized["task_queue_name"]
            ] = self.object_manager.dict()

        return super().queue(serialized)



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import datetime
from dataclasses import dataclass, field
from functools import cached_property

import ulid

from quesadilla.core.task_executables import TaskExecutionContext, TaskExecutionResult
from quesadilla.core.task_serialization import (
    SerializedTask,
    from_timestamp,
    from_timestamp_or_none,
    to_timestamp,
    to_timestamp_or_none,
)


@dataclass(frozen=True)
class BaseTaskMetadata[T]:
    execution_context: TaskExecutionContext = field(repr=False)
    execution_result: TaskExecutionResult[T] = field(
        default_factory=TaskExecutionResult
    )

    id: ulid.ULID = field(default_factory=ulid.ULID)

    queued_at: datetime.datetime = field(default=datetime.datetime.now(datetime.UTC))
    started_at: datetime.datetime | None = None
    finalized_at: datetime.datetime | None = None

    @cached_property
    def started(self) -> bool:
        return self.started_at is not None

    @cached_property
    def finalized(self) -> bool:
        return self.finalized_at is not None


@dataclass(frozen=True)
class SerializableTask[T](BaseTaskMetadata[T]):
    namespace: str = field(kw_only=True)
    task_queue_name: str = field(kw_only=True)
    task_name: str = field(kw_only=True)

    def serialize(self) -> SerializedTask[T]:
        return {
            "id": self.id,
            "namespace": self.namespace,
            "task_queue_name": self.task_queue_name,
            "task_name": self.task_name,
            "execution_context": {
                "args": self.execution_context.args,
                "kwargs": self.execution_context.kwargs,
            },
            "execution_result": {
                "finished": self.execution_result.finished,
                "value": self.execution_result.value,
                "exc": self.execution_result.exc,
            },
            "queued_at": to_timestamp(self.queued_at),
            "started_at": to_timestamp_or_none(self.started_at),
            "finalized_at": to_timestamp_or_none(self.finalized_at),
        }

    @staticmethod
    def deserialize[NT](serialized: SerializedTask[NT]) -> SerializableTask[NT]:
        return SerializableTask(
            id=serialized["id"],
            namespace=serialized["namespace"],
            task_queue_name=serialized["task_queue_name"],
            task_name=serialized["task_name"],
            execution_context=TaskExecutionContext(
                args=serialized["execution_context"]["args"],
                kwargs=serialized["execution_context"]["kwargs"],
            ),
            execution_result=TaskExecutionResult(
                finished=serialized["execution_result"]["finished"],
                value=serialized["execution_result"]["value"],
                exc=serialized["execution_result"]["exc"],
            ),
            queued_at=from_timestamp(serialized["queued_at"]),
            started_at=from_timestamp_or_none(serialized["started_at"]),
            finalized_at=from_timestamp_or_none(serialized["finalized_at"]),
        )



# `quesadilla` - an elegant background task queue for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TypedDict


class TaskQueueLoggingExtras(TypedDict):
    namespace: str
    task_queue_name: str


class TaskLoggingExtras(TypedDict):
    task_queue: TaskQueueLoggingExtras
    task_name: str


class QueuedTaskLoggingExtras(TaskLoggingExtras):
    id: str
    finished: bool
